package net.knarcraft.minecraftserverlauncher.utility;

import net.knarcraft.minecraftserverlauncher.Main;
import net.knarcraft.minecraftserverlauncher.profile.Collection;
import net.knarcraft.minecraftserverlauncher.server.Server;
import net.knarcraft.minecraftserverlauncher.userinterface.BackupGUI;
import net.knarcraft.minecraftserverlauncher.userinterface.GUI;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;
import java.util.stream.Stream;

/**
 * A helper class for performing server backup
 */
public class BackupUtil {

    private static boolean backupAborted;
    private static boolean backupRunning = false;

    /**
     * Aborts the currently running backup
     */
    public static void abortBackup() {
        backupAborted = true;
    }

    /**
     * Recursively copies a folder to another location
     *
     * @param source      <p>The folder to copy</p>
     * @param destination <p>Target destination</p>
     * @throws IOException <p>If we can't start a file stream</p>
     */
    private static synchronized long backupFolder(File source, File destination, long backupFileSize, long alreadyCopied) throws IOException {
        if (backupAborted) {
            return 0L;
        }
        if (!source.isDirectory()) {
            long copiedFileSize = copyFile(source, destination);
            BackupGUI.updateProgress("Copying " + source + "\n to " + destination,
                    (int)((alreadyCopied + copiedFileSize) * 100 / backupFileSize));
            return copiedFileSize;
        } else {
            if (!destination.exists() && !destination.mkdir()) {
                return 0L;
            }
            String[] files = source.list();
            long copiedFilesSize = 0;
            if (files != null) {
                for (String file : files) {
                    File srcFile = new File(source, file);
                    File destinationFile = new File(destination, file);
                    copiedFilesSize += backupFolder(srcFile, destinationFile, backupFileSize,
                            alreadyCopied + copiedFilesSize);
                    BackupGUI.updateProgress("Copying " + source + "\n to " + destination,
                            (int)((alreadyCopied + copiedFilesSize) * 100 / backupFileSize));
                }
            }
            return copiedFilesSize;
        }
    }

    /**
     * Copies a file from one location to another
     *
     * @param source            <p>The file to copy</p>
     * @param destination       <p>The location of the copied file</p>
     * @throws IOException <p>If reading or writing fails</p>
     */
    private static long copyFile(File source, File destination) throws IOException {
        InputStream in = new FileInputStream(source);
        OutputStream out = new FileOutputStream(destination);
        byte[] buffer = new byte[1024];
        int length;
        while ((length = in.read(buffer)) > 0) {
            out.write(buffer, 0, length);
        }
        in.close();
        out.close();
        return Files.size(source.toPath());
    }

    /**
     * Copies all server directories to a folder specified by the user
     */
    public synchronized static void backup(GUI gui) {
        backupAborted = false;
        if (backupRunning) {
            gui.setStatus("A backup is already running");
            return;
        } else {
            backupRunning = true;
        }

        //Get the folder to save the backed up files in
        File path = gui.askForDirectory("Backup folder");
        if (path == null || !path.isDirectory()) {
            backupRunning = false;
            return;
        }

        gui.setStatus("Backup running...");

        List<List<File>> serverFolders = getFoldersOfEnabledServers(path);

        long backupFileSize = getFolderSize(gui, serverFolders);

        gui.setStatus("Backing up " + (backupFileSize / 1000000) + "MB");

        new BackupGUI();
        BackupGUI.updateProgress("Backup starting...", 0);
        long alreadyCopied = 0;
        for (List<File> serverFolder : serverFolders) {
            if (backupAborted) {
                gui.setStatus("Backup aborted");
                backupRunning = false;
                return;
            }
            File srcFolder = serverFolder.get(0);
            File destinationFolder = serverFolder.get(1);

            //Create child folders
            if (!destinationFolder.exists() && !destinationFolder.mkdirs()) {
                backupRunning = false;
                gui.logError("Unable to create necessary sub-folder in the backup folder");
                throw new IllegalArgumentException("Unable to create necessary sub-folder in the backup folder");
            }
            //Backup
            try {
                alreadyCopied += backupFolder(srcFolder, destinationFolder, backupFileSize, alreadyCopied);
            } catch (IOException e) {
                e.printStackTrace();
                gui.setStatus("Backup caused an error");
                backupRunning = false;
            }
        }
        backupRunning = false;
        BackupGUI.destroy();
        if (backupAborted) {
            gui.setStatus("Backup aborted");
        } else {
            gui.setStatus("Backup finished");
        }
    }

    /**
     * Gets the size of a list of folders
     *
     * @param gui <p>The GUI to write any errors to</p>
     * @param serverFolders <p>The folder to find the size of</p>
     * @return <p>The size of the given folders</p>
     */
    private static long getFolderSize(GUI gui, List<List<File>> serverFolders) {
        long folderSize = 0;
        for (List<File> serverFolder : serverFolders) {
            File srcFolder = serverFolder.get(0);
            try (Stream<Path> walk = Files.walk(srcFolder.toPath())) {
                folderSize += walk.filter(Files::isRegularFile).mapToLong(BackupUtil::getFileSize).sum();
            } catch (IOException e) {
                gui.setStatus(String.format("IO errors %s", e));
            }
        }
        return folderSize;
    }

    /**
     * Gets the input and output folders for enabled servers
     *
     * <p>The input folders are the folders to copy, while the output folders are the folders to write the backup to.
     * Each list element contains a list of exactly two File items. The first one is the input folder, and the second
     * one is the output folder</p>
     *
     * @param path <p>The path of the backup folder, as given by the user</p>
     * @return <p>The folders to copy from/to</p>
     */
    private static List<List<File>> getFoldersOfEnabledServers(File path) {
        List<List<File>> serverFolders = new ArrayList<>();

        //Get folders of servers to back up
        List<Collection> collections = Main.getController().getCurrentProfile().getCollections();
        for (Collection collection : collections) {
            //Ignore disabled and invalid servers
            if (collection.getServer().getPath().equals("") || !collection.getServer().isEnabled()) {
                continue;
            }
            //Decide sub-folders
            Server targetServer = collection.getServer();
            String name = targetServer.getName();
            File srcFolder = new File(targetServer.getPath());
            File destinationFolder = new File(path, name);

            List<File> serverFolder = new ArrayList<>();
            serverFolder.add(srcFolder);
            serverFolder.add(destinationFolder);
            serverFolders.add(serverFolder);
        }
        return serverFolders;
    }

    /**
     * Gets the size of a file given its path
     * @param path <p>The path to a file</p>
     * @return <p>The size of the file in bytes, or 0 if an exception is thrown</p>
     */
    private static long getFileSize(Path path) {
        try {
            return Files.size(path);
        } catch (IOException exception) {
            System.out.printf("Failed to get size of %s%n%s", path, exception);
            return 0L;
        }
    }

}
